@@ -1,6 +1,6 @@
 // This is the implementation of the Chimera class.
 //
-// Copyright (c) 2021 Riverbank Computing Limited <info@riverbankcomputing.com>
+// Copyright (c) 2024 Riverbank Computing Limited <info@riverbankcomputing.com>
 // 
 // This file is part of PyQt6.
 // 
@@ -33,7 +33,7 @@
 
 
 // The registered user-defined Python enum types.
-QSet<PyObject *> Chimera::_py_enum_types;
+QHash<PyObject *, QByteArray> Chimera::_py_enum_types;
 
 // The cache of previously parsed argument type lists.
 QHash<QByteArray, QList<const Chimera *> > Chimera::_previously_parsed;
@@ -129,7 +129,8 @@ static MappedTypeConvertor<QVariantPair> variant_pair(
 
 // Construct an invalid type.
 Chimera::Chimera()
-    : _type(0), _py_type(0), _inexact(false), _is_qflags(false)
+    : _type(0), _py_type(0), _inexact(false), _is_qflags(false),
+      _is_char_star(false)
 {
 }
 
@@ -146,6 +147,7 @@ Chimera::Chimera(const Chimera &other)
     _inexact = other._inexact;
     _is_qflags = other._is_qflags;
     _name = other._name;
+    _is_char_star = other._is_char_star;
 }
 
 
@@ -156,11 +158,13 @@ Chimera::~Chimera()
 }
 
 
-// Register the type of a user-defined Python enum.
-void Chimera::registerPyEnum(PyObject *enum_type)
+// Register the type and pseudo fully qualified C++ name of a user-defined
+// Python enum.
+void Chimera::registerPyEnum(PyObject *enum_type,
+        const QByteArray &cpp_qualname)
 {
     Py_INCREF(enum_type);
-    _py_enum_types.insert(enum_type);
+    _py_enum_types.insert(enum_type, cpp_qualname);
 }
 
 
@@ -246,6 +250,17 @@ const Chimera *Chimera::parse(const QMetaProperty &mprop)
 
         if (menum.scope())
             type_name.prepend("::").prepend(menum.scope());
+
+        // If it is a QFlag then we have the typedef name rather than the
+        // underlying type which SIP knows about, therefore resolve the typedef
+        // name.
+        if (mprop.isFlagType())
+        {
+            const char *resolved = sipResolveTypedef(type_name.constData());
+
+            if (resolved)
+                type_name = QByteArray(resolved);
+        }
     }
     else
     {
@@ -469,6 +484,10 @@ bool Chimera::parse_py_type(PyTypeObject *type_obj)
         if (sipTypeIsEnum(td))
         {
             metatype = QMetaType::fromName(_name);
+
+            // Unregistered enums are handled as integers.
+            if (!metatype.isValid())
+                metatype = QMetaType(QMetaType::Int);
         }
         else if (_is_qflags)
         {
@@ -505,7 +524,9 @@ bool Chimera::parse_py_type(PyTypeObject *type_obj)
     else if (_py_enum_types.contains((PyObject *)type_obj))
     {
         metatype = QMetaType(QMetaType::Int);
-        _name = sipPyTypeName(type_obj);
+
+        // Note that we use the fully qualified name.
+        _name = _py_enum_types.value((PyObject *)type_obj);
     }
     else if (type_obj == &PyList_Type)
     {
@@ -540,7 +561,7 @@ bool Chimera::parse_py_type(PyTypeObject *type_obj)
     }
 
     // Fallback to using a PyQt_PyObject.
-    if (metatype.id() == QMetaType::UnknownType)
+    if (!metatype.isValid())
         metatype = QMetaType::fromType<PyQt_PyObject>();
 
     // If there is no name so far then use the meta-type name.
@@ -698,8 +719,10 @@ bool Chimera::parse_cpp_type(const QByteArray &type)
     // See if the type is known to Qt.
     metatype = QMetaType::fromName(resolved);
 
-    // If not then use the PyQt_PyObject wrapper.
-    if (!metatype.isValid())
+    // If not then use the PyQt_PyObject wrapper for now.
+    bool valid = metatype.isValid();
+
+    if (!valid)
         metatype = QMetaType::fromType<PyQt_PyObject>();
 
     // See if the type (without a pointer) is known to SIP.
@@ -732,6 +755,14 @@ bool Chimera::parse_cpp_type(const QByteArray &type)
         if (metatype != QMetaType::fromType<PyQt_PyObject>() && !is_ptr)
             return true;
 
+        if (resolved == "char" && is_ptr)
+        {
+            // This is (hopefully) '\0' terminated string.
+            _is_char_star = true;
+
+            return true;
+        }
+
         // This is an explicit 'PyQt_PyObject'.
         if (resolved == "PyQt_PyObject" && !is_ptr)
             return true;
@@ -744,7 +775,8 @@ bool Chimera::parse_cpp_type(const QByteArray &type)
 
     set_qflags(_type);
 
-    if (_is_qflags)
+    // We always pass these as integers.
+    if ((!valid && sipTypeIsEnum(_type)) || _is_qflags)
         metatype = QMetaType(QMetaType::Int);
 
     if (is_ptr)
@@ -786,7 +818,7 @@ bool Chimera::fromPyObject(PyObject *py, void *cpp) const
 
     PyErr_Clear();
 
-    switch (metatype.id())
+    switch (typeId())
     {
     case QMetaType::Nullptr:
         *reinterpret_cast<std::nullptr_t *>(cpp) = nullptr;
@@ -797,8 +829,15 @@ bool Chimera::fromPyObject(PyObject *py, void *cpp) const
         break;
 
     case QMetaType::Int:
-        *reinterpret_cast<int *>(cpp) = PyLong_AsLong(py);
-        break;
+        {
+            if (isEnumOrFlag())
+                *reinterpret_cast<QVariant *>(cpp) = QVariant(
+                        get_enum_value(py));
+            else
+                *reinterpret_cast<int *>(cpp) = PyLong_AsLong(py);
+
+            break;
+        }
 
     case QMetaType::UInt:
         *reinterpret_cast<unsigned int *>(cpp) = sipLong_AsUnsignedLong(py);
@@ -931,6 +970,20 @@ bool Chimera::fromPyObject(PyObject *py, void *cpp) const
             break;
         }
 
+    case -1:
+        {
+            const char **ptr = reinterpret_cast<const char **>(cpp);
+
+            if (PyBytes_Check(py))
+                *ptr = PyBytes_AsString(py);
+            else if (py == Py_None)
+                *ptr = 0;
+            else
+                iserr = 1;
+
+            break;
+        }
+
     default:
         if (_type)
         {
@@ -1063,7 +1116,7 @@ bool Chimera::fromPyObject(PyObject *py, QVariant *var, bool strict) const
     QVariant variant;
     int metatype_used = metatype.id();
 
-    switch (metatype.id())
+    switch (typeId())
     {
     case QMetaType::Nullptr:
         tmp_storage.tmp_nullptr_t = nullptr;
@@ -1074,7 +1127,11 @@ bool Chimera::fromPyObject(PyObject *py, QVariant *var, bool strict) const
         break;
 
     case QMetaType::Int:
-        if (_inexact)
+        if (isEnumOrFlag())
+        {
+            tmp_storage.tmp_int = get_enum_value(py);
+        }
+        else if (_inexact)
         {
             // Fit it into the smallest C++ type we can.
 
@@ -1278,11 +1335,24 @@ bool Chimera::fromPyObject(PyObject *py, QVariant *var, bool strict) const
             break;
         }
 
+    case -1:
+        metatype_used = QMetaType::VoidStar;
+
+        if (PyBytes_Check(py))
+            tmp_storage.tmp_void_ptr = PyBytes_AsString(py);
+        else if (py == Py_None)
+            tmp_storage.tmp_void_ptr = 0;
+        else
+            iserr = 1;
+
+        break;
+
     default:
         if (_type)
         {
             if (sipTypeIsEnum(_type))
             {
+                // Note that this will be a registered enum.
                 tmp_storage.tmp_int = sipConvertToEnum(py, _type);
             }
             else if (_name.endsWith('*'))
@@ -1405,11 +1475,17 @@ PyObject *Chimera::toPyObject(const QVariant &var) const
         // A sanity check.
         if (metatype != var.metaType())
         {
-            PyErr_Format(PyExc_TypeError,
-                    "unable to convert a QVariant of type %d to a QMetaType of type %d",
-                    var.metaType().id(), metatype.id());
+            // However having an Int QVariant value and having an invalid
+            // QMetaType is the case that we are dealing with a user-defined
+            // enum and so is valid.
+            if (!(var.metaType().id() == QMetaType::Int && metatype.id() == QMetaType::UnknownType))
+            {
+                PyErr_Format(PyExc_TypeError,
+                        "unable to convert a QVariant of type %d to a QMetaType of type %d",
+                        var.metaType().id(), metatype.id());
 
-            return 0;
+                return 0;
+            }
         }
 
         // Deal with the simple case of unwrapping a Python object rather than
@@ -1450,38 +1526,50 @@ PyObject *Chimera::toPyObject(void *cpp) const
 {
     if (metatype == QMetaType::fromType<PyQt_PyObject>())
     {
-        if (_type)
+        if (_is_char_star)
+        {
+            char *s = *reinterpret_cast<char **>(cpp);
+
+            if (s)
+                return PyBytes_FromString(s);
+
+            Py_INCREF(Py_None);
+            return Py_None;
+        }
+
+        if (_type && !sipTypeIsEnum(_type))
         {
-            // SIP knows the type (therefore it isn't really wrapped in a
-            // PyQt_PyObject) but it's not registered with Qt.
             if (_name.endsWith('*'))
                 cpp = *reinterpret_cast<void **>(cpp);
 
             return sipConvertFromType(cpp, _type, 0);
         }
-        else
+
+        // Unwrap a Python object.
+        PyQt_PyObject *pyobj_wrapper = reinterpret_cast<PyQt_PyObject *>(cpp);
+
+        if (!pyobj_wrapper->pyobject)
         {
-            // Otherwise unwrap the Python object.
-            PyQt_PyObject *pyobj_wrapper = reinterpret_cast<PyQt_PyObject *>(cpp);
+            PyErr_SetString(PyExc_TypeError,
+                    "unable to convert a QVariant back to a Python object");
 
-            if (!pyobj_wrapper->pyobject)
-            {
-                PyErr_SetString(PyExc_TypeError,
-                        "unable to convert a QVariant back to a Python object");
-
-                return 0;
-            }
-
-            Py_INCREF(pyobj_wrapper->pyobject);
-
-            return pyobj_wrapper->pyobject;
+            return 0;
         }
+
+        Py_INCREF(pyobj_wrapper->pyobject);
+
+        return pyobj_wrapper->pyobject;
     }
 
     PyObject *py = 0;
 
     switch (metatype.id())
     {
+    case QMetaType::UnknownType:
+        // In invalid QMetaType probably means a user-defined enum.
+        py = get_enum_key(*reinterpret_cast<int *>(cpp));
+        break;
+
     case QMetaType::Nullptr:
         py = Py_None;
         Py_INCREF(py);
@@ -1496,6 +1584,10 @@ PyObject *Chimera::toPyObject(void *cpp) const
         {
             py = sipConvertFromType(cpp, _type, 0);
         }
+        else if (isEnumOrFlag())
+        {
+            py = get_enum_key(*reinterpret_cast<int *>(cpp));
+        }
         else
         {
             py = PyLong_FromLong(*reinterpret_cast<int *>(cpp));
@@ -1592,6 +1684,7 @@ PyObject *Chimera::toPyObject(void *cpp) const
         {
             if (sipTypeIsEnum(_type))
             {
+                // Note that this will be a registered enum.
                 py = sipConvertFromEnum(*reinterpret_cast<int *>(cpp), _type);
             }
             else if (_name.endsWith('*'))
@@ -1737,6 +1830,59 @@ bool Chimera::isEnumOrFlag() const
 }
 
 
+// Return the value of an enum, either a wrapped one or a user-defined one.
+int Chimera::get_enum_value(PyObject *py) const
+{
+    int value = 0;
+
+    if (_type)
+    {
+        value = sipConvertToEnum(py, _type);
+    }
+    else
+    {
+        PyObject *value_obj = PyObject_GetAttrString(py, "value");
+
+        if (value_obj)
+        {
+            value = PyLong_AsLong(value_obj);
+            Py_DECREF(value_obj);
+        }
+    }
+
+    return value;
+}
+
+
+// Return the key of an enum, either a wrapped one or a user-defined one.
+PyObject *Chimera::get_enum_key(int cpp) const
+{
+    PyObject *py = NULL;
+
+    if (_type)
+    {
+        py = sipConvertFromEnum(cpp, _type);
+    }
+    else
+    {
+        QHashIterator<PyObject *, QByteArray> it(_py_enum_types);
+
+        while (it.hasNext())
+        {
+            it.next();
+
+            if (it.value() == _name)
+            {
+                py = PyObject_CallFunction(it.key(), "(i)", cpp);
+                break;
+            }
+        }
+    }
+
+    return py;
+}
+
+
 // Convert a Python object to C++, allocating storage as necessary.
 Chimera::Storage *Chimera::fromPyObjectToStorage(PyObject *py) const
 {
